#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Deployment status data model."""

import asyncio
import copy
import dataclasses
import enum
import logging
import typing

import click

log = logging.getLogger(__name__)

DEFAULT_MONITOR_SLEEP_SECONDS = 10


@dataclasses.dataclass
class DeploymentState:
    """Signal state of a deployment."""

    @enum.unique
    class ResultType(enum.Enum):
        """Deployment state enumeration."""

        cancelled = enum.auto()
        failed = enum.auto()
        in_progress = enum.auto()
        pending = enum.auto()
        success = enum.auto()

    code: ResultType
    message: typing.Optional[str] = None


T = typing.TypeVar("T", bound="DeploymentStatus")


class DeploymentStatus:
    """Coordinate reporting of asynchronous deployment status."""

    __iteration_context: str
    __lock: asyncio.Lock
    __status: typing.Dict[str, DeploymentState]

    STATE_COLOURS = {
        DeploymentState.ResultType.pending: "white",
        DeploymentState.ResultType.in_progress: "cyan",
        DeploymentState.ResultType.cancelled: "yellow",
        DeploymentState.ResultType.failed: "red",
        DeploymentState.ResultType.success: "green",
    }

    def __init__(self: T, iteration_context: str) -> None:
        """Construct ``DeploymentStatus`` object."""
        self.__iteration_context = iteration_context
        self.__lock = asyncio.Lock()
        self.__status = dict()

    async def initialize(self: T, name: str) -> None:
        """
        Register a new deployment name to record status against.

        Args:
            name: Name of deployment.

        """
        if name in self.__status:
            log.warning(
                "Re-initializing existing status entry, {0}".format(name)
            )
        async with self.__lock:
            self.__status[name] = DeploymentState(
                code=DeploymentState.ResultType.pending
            )

    async def write(
        self: T,
        name: str,
        code: DeploymentState.ResultType,
        message: typing.Optional[str] = None,
    ) -> None:
        """
        Write deployment state for specified deployment name.

        Args:
            name: Name of deployment.
            code: Status code to record.
            message: Status message (optional).
        """
        async with self.__lock:
            self.__status[name].code = code
            self.__status[name].message = message

    async def read(self: T, name: str) -> DeploymentState:
        """
        Read deployment state for specified deployment name.

        Args:
            name: Name of deployment.

        Returns:
            Read state.
        Raises:
            KeyError: If name does not exist.
        """
        async with self.__lock:
            result = copy.deepcopy(self.__status[name])

        return result

    async def names(self: T) -> typing.Set[str]:
        """
        Read the name registered for deployment monitoring.

        Returns:
            Set of deployment names registered.
        """
        async with self.__lock:
            result = set(self.__status.keys())

        return result

    async def __monitor_status(self: T) -> None:
        completed = False
        while not completed:
            status = {n: await self.read(n) for n in await self.names()}
            if not status:
                message = "nothing to report, {0}".format(
                    self.__iteration_context
                )
                log.info(message)
                click.echo(click.style(message, fg="yellow"))
            else:
                for n, s in status.items():
                    message = "{0}: {1} {2}".format(
                        self.__iteration_context, n, s.code.name
                    )
                    this_colour = self.STATE_COLOURS[s.code]
                    log.info(message)
                    click.echo(click.style(message, fg=this_colour))
                if all(
                    [
                        x
                        in [
                            DeploymentState.ResultType.success,
                        ]
                        for x in status.values()
                    ]
                ):
                    completed = True
                    log.info(
                        "all deployments succeeded for context, {0}".format(
                            self.__iteration_context
                        )
                    )
                elif all(
                    [
                        x
                        in [
                            DeploymentState.ResultType.success,
                            DeploymentState.ResultType.failed,
                            DeploymentState.ResultType.cancelled,
                        ]
                        for x in status.values()
                    ]
                ):
                    completed = True
                    log.info(
                        "all deployments completed with some failures for "
                        "context, {0}".format(self.__iteration_context)
                    )

            if not completed:
                log.debug(
                    "status monitor sleeping for, {0} seconds".format(
                        DEFAULT_MONITOR_SLEEP_SECONDS
                    )
                )
                await asyncio.sleep(DEFAULT_MONITOR_SLEEP_SECONDS)

    async def start_monitor(self: T) -> None:
        """
        Start the concurrent deployment status monitor.

        Status is reported to console and log until all members have hit one
        of the three deployment termination states [success|failed|cancelled].
        """
        asyncio.create_task(self.__monitor_status())



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio

import pytest

from foodx_devops_tools.deploy_me._deployment import (
    DeploymentState,
    DeploymentStatus,
    PipelineCliOptions,
    deploy_frame,
)
from foodx_devops_tools.deploy_me.exceptions import DeploymentCancelledError
from foodx_devops_tools.pipeline_config import IterationContext

MOCK_ITERATION_CONTEXT = IterationContext()
MOCK_ITERATION_CONTEXT.append("some")
MOCK_ITERATION_CONTEXT.append("context")


@pytest.fixture()
def mock_application_deploy(mock_async_method, prep_frame_data):
    deployment_data, frame_data = prep_frame_data
    mock_application = mock_async_method(
        "foodx_devops_tools.deploy_me._deployment.deploy_application"
    )

    return mock_application, deployment_data, frame_data


@pytest.mark.asyncio
async def test_validation_clean(mocker, mock_application_deploy):
    cli_options = PipelineCliOptions(
        enable_validation=True,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=(15 * 60),
    )

    mock_application, deployment_data, frame_data = mock_application_deploy

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )

    mock_application.assert_called_once_with(
        frame_data.applications["a1"],
        mocker.ANY,
        mocker.ANY,
        cli_options.enable_validation,
        frame_data.folder,
    )


@pytest.mark.asyncio
async def test_dependency_failed(mock_application_deploy):
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await this_status.initialize(dependency_frame)
    await this_status.write(dependency_frame, DeploymentState.ResultType.failed)

    await check_is_cancelled(this_status, frame_data, deployment_data)


async def check_is_cancelled(this_status, frame_data, deployment_data):
    cli_options = PipelineCliOptions(
        enable_validation=True,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=(15 * 60),
    )
    with pytest.raises(
        DeploymentCancelledError, match=r"^cancelled due to dependency failure"
    ):
        await deploy_frame(
            frame_data,
            deployment_data,
            this_status,
            cli_options,
        )


@pytest.mark.asyncio
async def test_dependency_cancelled(mock_application_deploy):
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await this_status.initialize(dependency_frame)
    await this_status.write(
        dependency_frame, DeploymentState.ResultType.cancelled
    )

    await check_is_cancelled(this_status, frame_data, deployment_data)


@pytest.mark.asyncio
async def test_multiple_dependencies_one_failed(mock_application_deploy):
    dependencies = {
        "df1": DeploymentState.ResultType.success,
        "df2": DeploymentState.ResultType.failed,
    }

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = list(dependencies.keys())

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    for this_df in dependencies.keys():
        await this_status.initialize(this_df)
        await this_status.write(this_df, dependencies[this_df])

    await check_is_cancelled(this_status, frame_data, deployment_data)


@pytest.mark.asyncio
async def test_dependencies_success(mock_application_deploy):
    cli_options = PipelineCliOptions(
        enable_validation=True,
        monitor_sleep_seconds=30,
        wait_timeout_seconds=(15 * 60),
    )
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await this_status.initialize(dependency_frame)
    await this_status.write(
        dependency_frame, DeploymentState.ResultType.success
    )

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )


async def delayed_completion(
    name: str,
    status: DeploymentStatus,
    duration_seconds: float,
    final_state: DeploymentState.ResultType,
):
    await status.write(name, DeploymentState.ResultType.in_progress)
    await asyncio.sleep(duration_seconds)
    await status.write(name, final_state)


@pytest.mark.asyncio
async def test_pausing(mock_application_deploy):
    cli_options = PipelineCliOptions(
        enable_validation=True, monitor_sleep_seconds=1, wait_timeout_seconds=2
    )
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await this_status.initialize(dependency_frame)

    await asyncio.create_task(
        delayed_completion(
            dependency_frame, this_status, 2, DeploymentState.ResultType.success
        )
    )

    await deploy_frame(
        frame_data,
        deployment_data,
        this_status,
        cli_options,
    )


@pytest.mark.asyncio
async def test_in_progress_timeout(mock_application_deploy):
    cli_options = PipelineCliOptions(
        enable_validation=True, monitor_sleep_seconds=2, wait_timeout_seconds=1
    )
    dependency_frame = "other-frame"

    mock_application, deployment_data, frame_data = mock_application_deploy
    frame_data.depends_on = [dependency_frame]

    this_status = DeploymentStatus(MOCK_ITERATION_CONTEXT)
    await this_status.initialize(dependency_frame)

    await asyncio.create_task(
        delayed_completion(
            dependency_frame,
            this_status,
            0.5,
            DeploymentState.ResultType.in_progress,
        )
    )

    with pytest.raises(
        DeploymentCancelledError, match=r"^timeout waiting for dependencies"
    ):
        await deploy_frame(
            frame_data,
            deployment_data,
            this_status,
            cli_options,
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""I/O related utilities."""

import typing


def acquire_token(password_file: typing.IO) -> str:
    """
    Read a password token from the specified stream.

    Args:
        password_file: Input file stream to read password token from.

    Returns:
        Read token.
    """
    with password_file:
        content = password_file.read()

    return content



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.deploy_me._status import (
    DeploymentState,
    DeploymentStatus,
)


@pytest.fixture()
def status_instance():
    _status_instance = DeploymentStatus("some.context")

    return _status_instance


class TestDeploymentStatus:
    @pytest.mark.asyncio
    async def test_clean(self, status_instance):
        this_name = "some_name"
        assert not await status_instance.names()
        await status_instance.initialize(this_name)

        assert await status_instance.names() == {this_name}

        await status_instance.write(
            this_name,
            DeploymentState.ResultType.pending,
            message="some message",
        )

        result = await status_instance.read(this_name)
        assert result.code == DeploymentState.ResultType.pending
        assert result.message == "some message"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Deployment status monitoring."""

import asyncio
import copy
import logging

import click

from foodx_devops_tools.pipeline_config import (
    IterationContext,
    SingularFrameDefinition,
)

from ._exceptions import DeploymentCancelledError
from ._state import PipelineCliOptions
from ._status import DeploymentState, DeploymentStatus

log = logging.getLogger(__name__)


async def _status_wait(
    this_context: IterationContext,
    frame_data: SingularFrameDefinition,
    frame_status: DeploymentStatus,
    status_monitor_sleep_seconds: float,
) -> None:
    assert frame_data.depends_on

    async def cancel_deployment() -> None:
        nonlocal frame_data, frame_status, this_context

        message = "cancelled due to dependency failure, {0}, {1}".format(
            this_context, str(frame_data.depends_on)
        )
        log.error(message)
        click.echo(click.style(message, fg="red"))
        await frame_status.write(
            str(this_context),
            DeploymentState.ResultType.cancelled,
            message=message,
        )
        raise DeploymentCancelledError(message)

    async def report_completion() -> None:
        nonlocal frame_status, this_context

        message = (
            "dependencies completed. proceeding with deployment, {0}".format(
                this_context
            )
        )
        log.info(message)
        click.echo(click.style(message, fg="cyan"))
        await frame_status.write(
            str(this_context), DeploymentState.ResultType.in_progress
        )

    async def report_missing_dependencies() -> None:
        nonlocal frame_data, frame_status, this_context

        message = "dependency frames not in frame status, {0}, {1}".format(
            this_context, str(frame_data.depends_on)
        )
        log.warning(message)
        click.echo(click.style(message, fg="yellow"))
        await frame_status.write(
            str(this_context),
            DeploymentState.ResultType.pending,
            message=message,
        )

    # pause until dependencies are complete
    paused = True
    while paused:
        try:
            dependency_context = list()
            base = copy.deepcopy(this_context)
            for x in frame_data.depends_on:
                base[-1] = x
                dependency_context.append(str(base))
            dependency_status = [
                (await frame_status.read(x)).code for x in dependency_context
            ]
            if any(
                [
                    x
                    in [
                        DeploymentState.ResultType.failed,
                        DeploymentState.ResultType.cancelled,
                    ]
                    for x in dependency_status
                ]
            ):
                await cancel_deployment()

            if all(
                [
                    x == DeploymentState.ResultType.success
                    for x in dependency_status
                ]
            ):
                await report_completion()
                paused = False
            else:
                message = "waiting for dependency completion, {0}, {1}".format(
                    this_context, dependency_context
                )
                log.info(message)
                click.echo(click.style(message))
                await asyncio.sleep(status_monitor_sleep_seconds)
        except KeyError:
            # we know the key is *supposed* to exist because it's defined in
            # `depends_on` (although it could be a typo), so keep trying in
            # case we just polled too early and it hasn't been registered
            # yet. the wait timeout will take care of it being the wrong value.
            await report_missing_dependencies()
            await asyncio.sleep(status_monitor_sleep_seconds)


async def process_dependencies(
    iteration_context: IterationContext,
    frame_data: SingularFrameDefinition,
    frame_status: DeploymentStatus,
    pipeline_parameters: PipelineCliOptions,
) -> None:
    """
    Wait for dependency frame status to succeed or fail.

    A timeout occurs if the specified duration is exceeded.

    Args:
        iteration_context: Current deployment hierarchy object.
        frame_data: Frame configuration data.
        frame_status: Status reporting object for frames.
        pipeline_parameters: Command line parameters.

    Raises:
        DeploymentCancelledError: If any dependencies don't complete or the
                                  maximum wait duration is exceeded.
    """
    # evaluate dependency status if they exist
    if frame_data.depends_on:
        try:
            await asyncio.wait_for(
                _status_wait(
                    iteration_context,
                    frame_data,
                    frame_status,
                    pipeline_parameters.monitor_sleep_seconds,
                ),
                timeout=pipeline_parameters.wait_timeout_seconds,
            )
        except asyncio.TimeoutError as e:
            message = "timeout waiting for dependencies, {0}".format(
                iteration_context
            )
            log.error(message)
            await frame_status.write(
                str(iteration_context),
                DeploymentState.ResultType.cancelled,
                message=message,
            )
            raise DeploymentCancelledError(message) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import io
import logging
import os
import pathlib
import tempfile
import typing

from foodx_devops_tools.pipeline_config import load_service_principals
from foodx_devops_tools.pipeline_config.service_principals import (
    _decrypt_vault,
    _encrypt_vault,
    managed_file_decrypt,
)


@contextlib.contextmanager
def encrypted_file(
    content: str, password: str
) -> typing.Generator[typing.Tuple[pathlib.Path, pathlib.Path], None, None]:
    with tempfile.TemporaryDirectory() as dir:
        dir_path = pathlib.Path(dir)
        encrypted_file_path = dir_path / "some.file"

        password_file_path = dir_path / "password-file"
        with password_file_path.open("w") as f:
            f.write(password)

        unencrypted_file_path = dir_path / "plain.file"
        with unencrypted_file_path.open("w") as f:
            f.write(content)

        _encrypt_vault(
            encrypted_file_path, password_file_path, unencrypted_file_path
        )

        os.remove(unencrypted_file_path)
        os.remove(password_file_path)

        assert encrypted_file_path.is_file()
        with encrypted_file_path.open("r") as f:
            encrypted_content = f.read()

        assert encrypted_content.startswith("$ANSIBLE_VAULT;1.1;AES256")

        yield encrypted_file_path


class TestManagedFile:
    def test_clean(self, caplog):
        with caplog.at_level(logging.DEBUG):
            file_text = """
    ---
    service_principals:
      sub1_name:
        id: 12345-id
        name: principal-name1
        secret: verysecret
    """
            decrypt_token = "somesecret"
            with encrypted_file(file_text, decrypt_token) as (
                encrypted_file_path
            ):
                expected_plain_file = pathlib.Path(
                    str(encrypted_file_path) + ".yml"
                )
                expected_password_file = pathlib.Path(
                    str(encrypted_file_path) + ".password"
                )
                # temporary files do not exist
                assert not expected_plain_file.exists()
                assert not expected_password_file.exists()

                with managed_file_decrypt(
                    encrypted_file_path, decrypt_token
                ) as f:
                    # plain text file has been created
                    assert expected_plain_file.is_file()

                    content = f.read()

                    assert content == file_text

                # temporary files have been erased
                assert not expected_plain_file.exists()
                assert not expected_password_file.exists()


class TestLoadServicePrincipals:
    mock_encrypted_path = pathlib.Path("some/file")
    mock_password_path = pathlib.Path("other/file")

    def test_simple(self, mocker):
        file_text = """
---
service_principals:
  sub1_name:
    id: 12345-id
    name: principal-name1
    secret: verysecret
  sub2_name:
    id: 123456-id
    name: principal-name2
    secret: verysecret
"""
        mock_stream = io.StringIO(initial_value=file_text)
        mock_context = mocker.patch(
            "foodx_devops_tools.pipeline_config.service_principals.managed_file_decrypt"
        )
        mock_context.return_value.__enter__.return_value = mock_stream

        result = load_service_principals(
            self.mock_encrypted_path, self.mock_password_path
        )

        assert len(result.service_principals) == 2
        assert result.service_principals["sub1_name"].id == "12345-id"
        assert result.service_principals["sub1_name"].name == "principal-name1"
        assert result.service_principals["sub1_name"].secret == "verysecret"

        assert result.service_principals["sub2_name"].id == "123456-id"

    def test_load(self):
        file_text = """
---
service_principals:
  sub1_name:
    id: 12345-id
    name: principal-name1
    secret: verysecret
  sub2_name:
    id: 123456-id
    name: principal-name2
    secret: verysecret
"""
        decrypt_token = "somesecret"
        with encrypted_file(file_text, decrypt_token) as (encrypted_file_path):
            result = load_service_principals(encrypted_file_path, decrypt_token)

        assert len(result.service_principals) == 2
        assert result.service_principals["sub1_name"].id == "12345-id"
        assert result.service_principals["sub1_name"].name == "principal-name1"
        assert result.service_principals["sub1_name"].secret == "verysecret"

        assert result.service_principals["sub2_name"].id == "123456-id"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging
import pathlib
import re
import typing

import click

from foodx_devops_tools.azure.cloud.auth import (
    AzureAuthenticationError,
    login_service_principal,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    AzureSubscriptionConfiguration,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    deploy as deploy_resource_group,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    validate as validate_resource_group,
)
from foodx_devops_tools.pipeline_config import (
    ApplicationDeploymentSteps,
    FlattenedDeployment,
    PipelineConfiguration,
    SingularFrameDefinition,
)

from ._dependency_monitor import process_dependencies
from ._state import PipelineCliOptions
from ._status import DeploymentState, DeploymentStatus

log = logging.getLogger(__name__)

SUBSCRIPTION_NAME_REGEX = (
    r"(?P<system>[a-z0-9]+)_(?P<client>[a-z0-9]+)_(?P<state>[a-z0-9]+)"
)


def assess_results(results: typing.List[DeploymentState]) -> DeploymentState:
    """
    Condense an array of deployment results into a single result.

    Args:
        results: Array of results to condense.

    Returns:
        Success, if all results are success. Failed otherwise.
    """
    if all([x.code == DeploymentState.ResultType.success for x in results]):
        this_result = DeploymentState(code=DeploymentState.ResultType.success)
    else:
        messages = [
            x.message
            for x in results
            if x.message and (x.code != DeploymentState.ResultType.success)
        ]
        this_result = DeploymentState(
            code=DeploymentState.ResultType.failed, message=str(messages)
        )

    return this_result


def _construct_resource_group_name(
    application_name: str, frame_name: str, client: str
) -> str:
    """Construct a resource group name from deployment context."""
    return "-".join([application_name, frame_name, client])


def _construct_fqdn(
    leaf_name: str, domain_root: str, client: str, subscription_name: str
) -> str:
    """Construct an FQDN from deployment context."""
    this_match = re.match(SUBSCRIPTION_NAME_REGEX, subscription_name)
    if this_match is not None:
        subd_state = this_match.group("state")
    else:
        # assume the entire subscription name is DNS valid
        subd_state = subscription_name

    return ".".join([leaf_name, subd_state, client, domain_root])


def _mangle_validation_resource_group(current_name: str, suffix: str) -> str:
    this_suffix = re.sub(r"[_.+]", "-", suffix)
    mangled_name = f"{current_name}-{this_suffix}"

    return mangled_name


async def deploy_application(
    application_data: ApplicationDeploymentSteps,
    deployment_data: FlattenedDeployment,
    application_status: DeploymentStatus,
    enable_validation: bool,
    frame_folder: pathlib.Path,
) -> None:
    """
    Deploy the steps of a frame application.

    Application steps must be deployed in sequence (serially).
    """
    this_context = str(deployment_data.data.iteration_context)

    try:
        message = "starting application deployment, {0}".format(this_context)
        log.info(message)
        click.echo(message)
        await application_status.initialize(this_context)

        await application_status.write(
            this_context, DeploymentState.ResultType.in_progress
        )

        puff_frame_data = deployment_data.data.puff_map.frames[
            deployment_data.context.frame_name
        ]
        puff_application_data = puff_frame_data.applications[
            deployment_data.context.application_name
        ]
        puff_parameter_data = puff_application_data.arm_parameters_files[
            deployment_data.context.release_state
        ][deployment_data.context.azure_subscription_name]

        this_subscription = AzureSubscriptionConfiguration(
            subscription_id=deployment_data.context.azure_subscription_name
        )
        for this_step in application_data:
            resource_group = (
                _construct_resource_group_name(
                    deployment_data.context.application_name,
                    deployment_data.context.frame_name,
                    deployment_data.context.client,
                )
                if not this_step.resource_group
                else this_step.resource_group
            )
            arm_template_path = (
                frame_folder / this_step.arm_file
                if this_step.arm_file
                else frame_folder
                / "{0}.json".format(deployment_data.context.application_name)
            )
            arm_parameters_path = (
                frame_folder / puff_parameter_data[this_step.name]
            )

            log.debug(
                "deployment_data.context, {0}, {1}".format(
                    this_context, str(deployment_data.context)
                )
            )
            log.debug(
                "deployment_data.data, {0}, {1}".format(
                    this_context, str(deployment_data.data)
                )
            )
            await login_service_principal(
                deployment_data.data.azure_credentials
            )
            if enable_validation:
                log.info(
                    "validation deployment enabled, {0}".format(this_context)
                )
                resource_group = _mangle_validation_resource_group(
                    resource_group,
                    deployment_data.context.pipeline_id,
                )
                log.info(
                    "validation resource group name, {0}".format(resource_group)
                )
                await validate_resource_group(
                    resource_group,
                    arm_template_path,
                    arm_parameters_path,
                    deployment_data.data.location_primary,
                    this_step.mode.value,
                    this_subscription,
                )
            else:
                log.info("deployment enabled, {0}".format(this_context))
                await deploy_resource_group(
                    resource_group,
                    arm_template_path,
                    arm_parameters_path,
                    deployment_data.data.location_primary,
                    this_step.mode.value,
                    this_subscription,
                )

        log.info("application deployment succeeded, {0}".format(this_context))
        await application_status.write(
            this_context, DeploymentState.ResultType.success
        )
    except asyncio.CancelledError:
        message = "async cancelled signal"
        log.error(message)
        await application_status.write(
            this_context,
            DeploymentState.ResultType.cancelled,
            message,
        )
        raise
    except AzureAuthenticationError as e:
        message = (
            "application deployment authentication "
            "failure, {0}, {1}, {2}".format(this_context, type(e), str(e))
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.error(message)
    except Exception as e:
        message = "application deployment failed, {0}, {1}, {2}".format(
            this_context, type(e), str(e)
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.exception(message)

    log.info("application deployment completed, {0}".format(this_context))


async def deploy_frame(
    frame_data: SingularFrameDefinition,
    deployment_data: FlattenedDeployment,
    frame_status: DeploymentStatus,
    pipeline_parameters: PipelineCliOptions,
) -> None:
    """
    Deploy the applications of a frame.

    Frame applications are deployed concurrently (in parallel).

    Raises:
        DeploymentCancelledError: If any dependencies fail preventing
                                  completion.
    """
    this_context = str(deployment_data.data.iteration_context)
    message = "starting frame deployment, {0}".format(this_context)
    log.info(message)
    click.echo(message)

    await frame_status.initialize(this_context)
    # application status will show as "pending" until deployment activates.
    application_status = DeploymentStatus(this_context)
    await application_status.start_monitor()

    await process_dependencies(
        deployment_data.data.iteration_context,
        frame_data,
        frame_status,
        pipeline_parameters,
    )

    await asyncio.gather(
        *[
            deploy_application(
                application_data,
                deployment_data.copy_add_application(application_name),
                application_status,
                pipeline_parameters.enable_validation,
                frame_data.folder,
            )
            for application_name, application_data in frame_data.applications.items()  # noqa: E501
        ],
        return_exceptions=False,
    )

    results = [await frame_status.read(x) for x in await frame_status.names()]
    condensed_result = assess_results(results)
    await frame_status.write(
        this_context, condensed_result.code, condensed_result.message
    )


async def do_deploy(
    configuration: PipelineConfiguration,
    deployment_data: FlattenedDeployment,
    pipeline_parameters: PipelineCliOptions,
) -> DeploymentState:
    """Deploy the frames in a flattened deployment configuration."""
    this_frames = configuration.frames
    deployment_data.data.iteration_context.append(
        deployment_data.data.deployment_tuple
    )
    deployment_data.data.puff_map = configuration.puff_map

    frame_deployment_status = DeploymentStatus(
        str(deployment_data.data.iteration_context)
    )
    await frame_deployment_status.start_monitor()

    await asyncio.gather(
        *[
            deploy_frame(
                frame_data,
                deployment_data.copy_add_frame(frame_name),
                frame_deployment_status,
                pipeline_parameters,
            )
            for frame_name, frame_data in this_frames.frames.items()
        ],
        return_exceptions=False,
    )

    results = [
        await frame_deployment_status.read(x)
        for x in await frame_deployment_status.names()
    ]
    condensed_result = assess_results(results)

    return condensed_result
